   /* IBM-PC-specific routines for SPIN */

#include "spin.h"
#include <dos.h>
#include <stdio.h>

/*****************************************************************************
SPIN handles the screen in the following way: it is assumed that both lines
and text can be drawn on the screen. Any screen has defined numbers of
character rows and columns, and defined numbers of x and y pels through
which lines may be drawn. SPIN finds these out through the enquiry function
getscreen(). (The screendef structure is defined in spin.h). Spin assumes
a cursor (not necessarily visible) at which the next character will be
positioned. SPIN expects to position the cursor with cursor(), to write
characters with prnc(), and to draw lines with line().

The following routines must be implemented on any computer which will run the
spin program. The present file is an implementation for IBM-PC or clones,
having VGA, EGA, CGA or Hercules compatible graphics board. For Hercules,
the HERCBIOS program must be run before SPIN.

void hardinit(void)             Do any required hardware initialization
                                This will be called first.

void getscreen(struct screendef *s) Fill in s with a description of the
                                screen which is available.

void hardfinish(void)           Tidy up hardware before program exit.

void cls(void)                  Clear the screen.

void cursor(int row, int col)   Place cursor at given row, col of screen,
                                0,0 is top left.

void prnc(char *s)              Display string s on screen at cursor posn
                                which is updated. Must handle \n and \r.

int key(void)                   Get a keystroke from user.

void statmsg(char *f, char *s)  Must do something sensible with the message
                                string s which has been emitted from the
                                function whose name is in f.

void errmsg(char *f, char *s)   As previous, for error messages. 

void zzstr(char *s, n)          Get a user string of max n-1 characters.

void line (int x1, int y1, int x2, int y2, unsigned char colour)
                                Draw a line from x1,y1 to x2,y2, in colour.
                                coordinates are in pels in 0,max where max is
                                known from getscreen(). 0,0 is top left

void screendump(int dotrows, char *file)
                                Do a graphics printer dump of the top dotrows 
                                rows, to the specified file. This can be a
                                dummy routine which just returns, if it's
                                impractical to implement.

End of required routines. Any routines below not in this list are secondary 
routines, for convenience, and will not be called directly by SPIN. Note that 
many of these routines are directly available in products such as TurboC. They 
have been implemented here for convenience in case they are not available with 
other compilers. The screendump() routine uses one TurboC specific routine, 
fileno(). This is flagged with a comment. Apart from this, the two required 
interface routines are int86(int n, union REGS *r, union REGS *r) which 
generates a software interrupt, and outportb(int port, char data) which writes 
an output port.  
*****************************************************************************/ 

char video(void);
void vmode(int m);
void vinit(void);
void pch(char c);
void setgraph(void);
void settext(void);
void bioprnx(char *s, int r, int c);
void scru(int t, int l, int b, int r, int n);
int crow(void);
int ccol(void);
void vline(int x, int y1, int y2);
void hline(int y, int x1, int x2);


#define BS         8       /* backspace on ASCII machine */
#define STATROW    24
#define STATCOL    60



static char board;
static int egavga, mxpel, mypel, colour;
static unsigned int VSEG = 0, GRAPHICS = 0, VIDMODE = 0xffff; 
static char __NATT = 7;


void zzstr(char *s, int n)     /* get string, allowing backspace corrections */
{  int r, c, i, k;

   i = 0;
   while (i < n-1)
   {   k = key();
       if (k >= ' ' && k <= '') s[i++] = k;
       if (k == BS && i != 0)
       {   pch(BS);
           pch(' ');
           pch(BS);
           i--;
       }
       if(k == '\n' || k == '\r') break;
   }
   s[i] = '\0';
}



void hardinit()
{
    settext();
    __NATT = 7;
    board = video();
    if (board == 'M' || board == '?')
    {   cls();
        bioprnx("This program needs a graphics board\n\n\n\r", 10, 20);
        exit(1);
    }
    egavga = ('E' == toupper(board) || 'V' == toupper(board));
    colour = (board == 'e' || board == 'v');

    mxpel = 639;
    if (board == 'H')
    {   mxpel = 719;
        mypel = 347;
    }
    else if (toupper(board) == 'E')
    {   mypel = 349;
        if (board == 'E') __NATT = 3;
    }
    else if (toupper(board) == 'V')
    {   mypel = 479;
    }
    else            /* Assume CGA (yecch) */
    {   mypel = 199;
    }
}


void hardfinish(void)
{
   settext();
}


void getscreen(struct screendef *s)
{
   s->ccols = 80;
   s->crows = (board == 'v' || board == 'V')? 30 : 25;
   s->xpels = mxpel + 1;
   s->ypels = mypel + 1;
}



void statmsg(char *f,  char *msg)
{
    bioprnx(msg, STATROW, STATCOL);
}


void errmsg(char *f,  char *msg)
{   settext();
    cursor(STATROW, 0);
    prnc("ERROR: ");
    prnc(msg);
    prnc("  in ");
    prnc(f);
    prnc("  ");
    key();
    exit(1);
}




void cursor(int row, int col)    /* move the cursor to row,col */
{   union REGS rr;
    rr.h.ah=15;
    int86(0x10,&rr,&rr);
    rr.h.dh=(row < 30 )? row : 29;
    rr.h.dl=(col < 80 )? col : 79;
    rr.h.ah=2;
    int86(0x10,&rr,&rr);
    return;
}



int crow(void)                  /* return current cursor row */
{   union REGS r;
    r.h.ah=15;
    int86(0x10,&r,&r);          /*gets active page to bh*/
    r.h.ah=3;
    int86(0x10,&r,&r);
    return((int)r.h.dh);
}




int ccol(void)                   /* return current cursor col */
{   union REGS r;
    r.h.ah=15;
    int86(0x10,&r,&r);          /*gets active page to bh*/
    r.h.ah=3;
    int86(0x10,&r,&r);
    return((int)r.h.dl);
}




void prnc(char *s)          /* print s at cursor using BIOS */
{   union REGS rr;
    int cr, cc;

    cr = crow();
    cc = ccol();
    rr.x.cx=1;
    rr.x.bx=__NATT;
    while ((rr.h.al=(*s++)) != 0)
    {   if (rr.h.al == '\r') cursor(cr, (cc=0));
        else if (rr.h.al == '\n')
        {   if (cr == 24) scru(0,0,24,79,1);
            else cr++;
            cursor(cr,cc);
        }
        else
        {   rr.h.ah=9;
            int86(0x10,&rr,&rr);
            cursor(cr,++cc);
        }
    }
    return;
}




void cls(void)                       /* Clear the Screen */
{   union REGS rr;
    rr.h.ah=15;
    int86(0x10,&rr,&rr);
    rr.h.ah=0;
    int86(0x10,&rr,&rr);
    return;
}





void pch(char c)                    /*put a char using BIOS (faster than */
                                    /*putch which calls DOS) */
{   union REGS r;
    r.h.al=c;
    r.h.bl=15;
    r.h.ah=14;
    int86(0x10,&r,&r);
    return;
}




void scru(int top, int left, int bot, int right, int n)
             /*scroll up defined rectangle n lines*/
{   union REGS r;

    if (0==VSEG) vinit();               /* set VIDMODE,VSEG if reqd. */

    r.h.al=n;
    r.h.ch=top;
    r.h.cl=left;
    r.h.dh=bot;
    r.h.dl=right;
    r.h.bh=7;
    r.h.ah=6;
    int86(0x10,&r,&r);
    return;
}




int key(void)                       /* read the keyboard */
{   int c,k;
    union REGS r;
    r.h.ah=0;
    int86(0x16,&r,&r);              /* read a key via BIOS which reveals all*/
    k=r.h.ah;
    c=r.h.al;

    if (c==0) c=256+k;          /* on IBM extended, return 256+ extension */

    if (c<256 && c>31) pch(c);

    return(c);
}



void vmode(int m)                /* set video mode */
{   union REGS r;
    r.h.ah=0;
    r.h.al=m;
    int86(0x10,&r,&r);
    vinit();
    return;
}




void vinit(void)          /* set VIDMODE, VSEG, GRAPHICS */

{   union REGS rr;
    char b;

    rr.h.ah=15;
    int86(0x10,&rr,&rr);
    VIDMODE=rr.h.al;
    b = video();
    if ((b == 'e' || b == 'v') && VIDMODE > 3) VSEG=0xA000;
    else if (VIDMODE < 7 ) VSEG=0xB800;
    else if (VIDMODE == 7) VSEG=0xB000;
    else if (VIDMODE == 0xF0) VSEG=0xB000;    /*Hercbios graphics mode*/
    else VSEG=0xA000;                         /*assume EGA otherwise*/

    GRAPHICS=(VIDMODE > 3) && (VIDMODE != 7);
    return;
}






char video(void)  /* Find out video board type */

              /* Returns 'E' if EGA with mono, 'e' for EGA with colour,
                'V' for mono vga, 'v' colour vga 'C' for CGA, 'M' for IBM mono,
                'H' for Hercules with HERCBIOS, '?' otherwise. */

{   union REGS r;
    char c1,c2='z',c3='a', far *vp, far *vp2;
    int info3;

    info3 = 0x08 & *(unsigned char far *)MK_FP(0x40, 0x87); /* ega active bit */

    r.x.ax=0xF000;
    int86(0x10,&r,&r);              /*HERCBIOS returns al= 'H'. ROM & EGA */
    if (r.h.al == 'H') return('H'); /* bios do nothing */

    r.x.ax = 0x1a00;
    int86(0x10,&r,&r);
    if (r.h.al == 0x1a)             /* a VGA */
    { if (r.h.bl == 1) return ('M');
      if (r.h.bl == 2) return ('C');
      if (r.h.bl == 4) return ('e');      /* hope Bruce Cowan is right! */
      if (r.h.bl == 5) return ('E');
      if (r.h.bl == 7) return ('V');
      if (r.h.bl == 8) return ('v');
      if (r.h.bl == 0xb) return ('V');    /* assume mcga behaves like vga ??? */
      if (r.h.bl == 0xc) return ('v');
    }
    r.h.ah=0x12;
    r.x.bx=0xFF10;
    int86(0x10,&r,&r);              /* EGA returns 0 or 1 in BH */
    if (r.h.bh == 1 && info3 == 0) return ('E');
    if (r.h.bh == 0 && info3 == 0) return ('e');
                                    /* drop through if EGA not active */
    r.h.ah=15;
    int86(0x10,&r,&r);              /* get video mode */
    if (r.h.al < 7) return ('C');   /* assume CGA if < 7 */

    if (r.h.al > 7) return ('?');   /* unknown if >7 & not EGA or Herc */

   vp = MK_FP(0xb000, 0);
   vp2 = MK_FP(0xb000, 0x2000);
   c1 = *vp;      /* save orig char */
   *vp = c2;      /* set VS:0, VS:2000 different */
   *vp2 = c3;
   c2 = *vp;      /* read VS:0; on IBM mono, address wrap makes it == VS:2000 */
   *vp = c1;      /* restore original character */

    if (c2 == c3) return ('M');
    else return ('?');
}




void setgraph(void)          /* put board into highest res graphics mode */
{
    char v;
    int m;
    v=video();
    if (v == 'H') m=0xf0;
    else if (v=='E') m=15;
    else if (v=='e') m=16;
    else if (v=='v' || v=='V') m = 0x12;
    else if (v=='C') m=6;
    else return;
    vmode(m);
    return;
}


void settext(void)          /* put board into text  mode */
{
    union REGS r;
    char v;
    int m;

    v=video();
    if (v=='e' || v=='v' ) m=3;
    else if (v=='C') m=2;
    else m=7;
    if (v == 'v' || v== 'V')    /* use 16 dot chars on vga */
    {   r.x.ax = 0x1202;
        r.h.bl = 0x30;
        int86(0x10, &r, &r);
    }
    vmode(m);
    return;
}




void bioprnx(char *s, int r, int c)    /* print s starting at r,c, using BIOS */
{
    cursor(r,c);
    prnc(s);
    return;
}





void line(int x1, int y1, int x2, int y2, unsigned char colour)
                         /* general line, screen dot coordinates */
{
    int dx, dy, d, st, i1, i2, x, y, xe, my1, my2, yi;

    if (!GRAPHICS) setgraph();

                    /* set ega map mask */
    if (board == 'E')
    {  if (colour < 8) colour = 3;
       else colour = 15;
    }
    if (egavga)
    {   outportb(0x3c4,2);
        outportb(0x3c5,colour);      /* mapmask */
    }


    if (y1 == y2)
    {   if (x2 > x1) hline(y1,x1,x2);
        else hline(y1,x2,x1);
        return;
    }
    if (x1 == x2)
    {   if (y2 > y1) vline(x1,y1,y2);
        else vline(x1,y2,y1);
        return;
    }

    dx = abs(x2-x1);    dy = abs(y2-y1);
    my1 = -y1;
    my2 = -y2;

    if (0 != (st = (dy > dx)))
    {   d = x1;     x1 = my1;    my1 = d;
        d = x2;     x2 = my2;    my2 = d;
        d = dx;     dx = dy;     dy = d;
    }
    i1 = dy + dy;
    d = i1 - dx;
    i2 = d - dx;
    yi = 1;

    if (x1 > x2)    { x = x2;   y = my2;     xe = x1;    if (my2>my1) yi = -1;}
    else            { x = x1;   y = my1;     xe = x2;    if (my1>my2) yi = -1;}

    while (x <= xe)
    {   if (st) wcdot(-x,y,colour);
        else wcdot(-y,x,colour);
        x++;
        if (d > 0)      { y += yi;      d += i2; }
        else d += i1;
    }
    return;
}



wcdot(row,col,colour)      /* write a graphics mode dot */
    register int col, row;
    char colour;
{
    unsigned int  roff;
    unsigned char pattern;
    static unsigned int step[4] = {0, 0x2000, 0x4000, 0x6000};

    if(egavga)  roff=((row<<3) + row + row) << 3;
    else if (board=='C') roff=80*(row>>1) + step[row & 1];
    else if (board=='H') roff=90*(row>>2) + step[row & 3];
    else return;

    pattern=(unsigned char)'\x80'>>(col&7);
    if (egavga) { outportb(0x3ce,8);    outportb(0x3cf,pattern);   /* bit mask */
                  outportb(0x3c4,2);    outportb(0x3c5,colour);    /* map mask */
                }
    *(unsigned char far *)MK_FP(VSEG, roff + (col>>3)) |= pattern;

    if(egavga)  { outportb(0x3ce,8);    outportb(0x3cf,0xff);   }  /* reset default*/
    
    return;
}

 

void vline(int col, int top, int bot)     /* vertical line in graphics mode */
{
    unsigned int  i, roff, tbyte, egavga;
    unsigned char far *vptr;
    unsigned char pattern;
    static char board=0;

    extern unsigned int VSEG, GRAPHICS;

    if (board == 0) board=video();
    if (VSEG == 0) vinit();
    if (!GRAPHICS) return;

    egavga = (board=='e' || board=='E' || board == 'v' || board == 'V');

    if(egavga)  roff=80*top;
    else if (board=='C') roff=80*(top>>1) + 0x2000*(top & 1);
    else if (board=='H') roff=90*(top>>2) + 0x2000*(top & 3);
    else return;

    pattern=(unsigned char)'\x80'>>(col&7);
    if (egavga)    { outportb(0x3ce,8);    outportb(0x3cf,pattern);    }
                                        /* bit mask */

    vptr=MK_FP(VSEG, roff + (col>>3));

    for (i=top;  i<=bot;  i++)
    {   *vptr |= pattern;
        if (egavga) vptr+=80;
        else if (board == 'C')
        {   if (0 == (i&1)) vptr+=8192;
            else vptr-=8112;
        }
        else                /* must be HERC */
        {   if (3 != (i&3)) vptr+=8192;
            else vptr-=24486;
        }
    }
    if(egavga)     { outportb(0x3ce,8);    outportb(0x3cf,0xff);   }   
                   /* reset default */
    return;
}




void hline(int row, int left, int right)  /* horizontal line in graphics mode */
{
    unsigned i,roff,lbyte,rbyte,nfill, egavga;
    unsigned char lpattern, rpattern;
    unsigned char far *vptr;
    static char board=0;

    extern unsigned int VSEG, GRAPHICS;

    if (board == 0) board=video();
    if (VSEG == 0) vinit();
    if (!GRAPHICS) return;

    egavga = (board=='e' || board=='E' || board=='v' || board=='V');


    if(egavga)  roff=80*row;
    else if (board=='C') roff=80*(row>>1) + 0x2000*(row & 1);
    else if (board=='H') roff=90*(row>>2) + 0x2000*(row & 3);
    else return;

    lbyte=roff + (left>>3);
    rbyte=roff + (right>>3);
    lpattern=(unsigned char)'\xff'>>(left & 7);
    rpattern=(unsigned char)'\xff'<<(7-(right & 7));

    vptr=MK_FP(VSEG,lbyte);
    if (lbyte==rbyte)
    {   lpattern &= rpattern;
        if(egavga) { outportb(0x3ce,8);    outportb(0x3cf,lpattern);   }
        (*vptr) |= lpattern;
        if(egavga) { outportb(0x3ce,8);    outportb(0x3cf,0xff);   }
        return;
    }
    else
    {   nfill=rbyte-lbyte-1;
        if(egavga) { outportb(0x3ce,8);    outportb(0x3cf,lpattern);   }
        (*vptr) |= lpattern;
        if(egavga) { outportb(0x3ce,8);    outportb(0x3cf,0xff);   }
        for (i=0;  i<nfill;  i++) (*++vptr) = '\xff';
        if(egavga) { outportb(0x3ce,8);    outportb(0x3cf,rpattern);   }
        (*++vptr) |= rpattern;
        if(egavga) { outportb(0x3ce,8);    outportb(0x3cf,0xff);   }
    }
    return;
}




#define MAXDOTS   720            /* maximum x pels, any board */
#define CRLF      "\x0a\x0d"
#define GRP       "\x1b*\x06"    /* Epson prefix for 90 dpi graphics */
#define GPERT     9              /* print graph spaces per text space */

void pgline(char *b, int ngc, FILE *f);


void screendump(int dotrows, char *name)  /* graphics screen dump of specified */
                  /* number of dot rows, starting at top of 
                     screen to file given by name */

{
    int  i, j, ibr, ibc, bytecols, byterows, handle;
    unsigned ngc, voff, b[8], r;
    register unsigned int s, q;
    unsigned char far *ptr;
    unsigned  char buf[MAXDOTS], *bp;
    FILE *f;
    union REGS rr;

    static unsigned char m[8]={'\x80', '\x40', '\x20', '\x10',
                               '\x08', '\x04', '\x02', '\x01'};

    static unsigned char emask[8]={'\xff', '\x80', '\xc0', '\xe0',
                                   '\xf0', '\xf8', '\xfc', '\xfe'};

    if (!GRAPHICS) return;

    f = fopen(name, "ab");
    if (f == NULL) 
    {  errmsg("screendump", "Can't open file or device");
       return;
    }

/* if a device, as opposed to a file, is specified for output, have to set
   DOS raw mode, otherwise an ASCII 26 will cause havoc. */

    handle = fileno(f);        /* Turbo - gets dos handle for a stream */
    rr.x.ax = 0x4400;
    rr.x.bx = handle;
    int86(0x21, &rr, &rr);     /* get device info */
    if (rr.x.dx & 0x80)        /* this bit will be set if it's a device */
    {  rr.x.ax = 0x4401;
       rr.x.bx = handle;
       rr.x.dx = 0x20;
       int86(0x21, &rr, &rr);  /* set device */
    }


    fputs("\x1bP\x1b\x33\x18\x12\x1bW\0", f);
         /* set printer pica, 1/9" line space , release compressed, expanded */

    if (board == 'H')    {   bytecols=90;       ngc=720;    }
    else                 {   bytecols=80;       ngc=640;    }

    if (colour)
    {   outportb(0x3ce,2);
        outportb(0x3cf,0);         /* colour compare reg = 0 */
        outportb(0x3ce,5);
        outportb(0x3cf,8);         /* read mode 1 */
    }

    byterows=(dotrows+7)>>3;

    for (ibr=0;  ibr<byterows;  ibr++)
    {   bp = buf;
        for (ibc=0;  ibc<bytecols;  ibc++)
        {   r=ibr<<3;
            if (egavga) voff=80*r;
            else if(board=='H') voff=0x2000*(r&3) + 90*(r>>2);
            else voff=0x2000*(r&1) + 80*(r>>1);
            ptr=MK_FP(VSEG,voff+ibc);

            for (j=0;  j<8;  j++)
            {
                b[j]=*ptr;
                if (egavga) ptr += 80; 
                else if (board == 'H') ptr += (3 == ((r+j)&3))? -24486 : 0x2000;
                else ptr += ((r+j)&1)? -8112 : 0x2000;
            }
            for (i=0;  i<8;  i++)
            {   s=0;
                q=m[i];
                if (b[0]&q) s |= '\x80';
                if (b[1]&q) s |= '\x40';
                if (b[2]&q) s |= '\x20';
                if (b[3]&q) s |= '\x10';
                if (b[4]&q) s |= '\x08';
                if (b[5]&q) s |= '\x04';
                if (b[6]&q) s |= '\x02';
                if (b[7]&q) s |= '\x01';

                if (colour) s^=0xff;
                if (ibr == byterows-1) s &= emask[dotrows&7];
                *bp++ = s;
            }
        }
        pgline(buf, ngc, f);
    }
    fputs("\x0d\x0a\x1b\x32\x0d\x0a", f);
    if (colour)  /* restore register defaults */
    {   outportb(0x3ce,5);
        outportb(0x3cf,0);  /* read mode 0 */
    }
    fclose(f);
    return;
}



void pgline(char *b, int ngc, FILE *f)/* print a graphics line using text spaces
                                                  when appropriate */
{
   int nb, ng, n, i, j;
   char *start;

   fputs(CRLF, f);
   nb = ng = 0;
   n = ngc - 1;
   start = b;

   for (i=0;  i<ngc;  i++)
   {  if (b[i] == 0) nb++;
      else
      {  ng += (nb + 1);
         nb = 0;
      }
      if (nb == GPERT || i == n)
      {  if (ng)
         {  fputs(GRP, f);
            fputc(ng & 255, f);
            fputc((ng >> 8) & 255, f);
            for (j =0;  j<ng;  j++) fputc(*(start + j), f);
            start += ng;
            ng = 0;
         }
         if (nb == GPERT)
         {  fputc(' ', f);
            start += GPERT;
            nb = 0;
         }
      }
   }
}



